package de.uni_marburg.powersort.benchmark;

import de.uni_marburg.powersort.data.DataEnum;
import de.uni_marburg.powersort.data.ObjectSupplier;
import org.openjdk.jmh.annotations.Benchmark;
import org.openjdk.jmh.annotations.BenchmarkMode;
import org.openjdk.jmh.annotations.Fork;
import org.openjdk.jmh.annotations.Level;
import org.openjdk.jmh.annotations.Measurement;
import org.openjdk.jmh.annotations.Mode;
import org.openjdk.jmh.annotations.OutputTimeUnit;
import org.openjdk.jmh.annotations.Param;
import org.openjdk.jmh.annotations.Scope;
import org.openjdk.jmh.annotations.Setup;
import org.openjdk.jmh.annotations.State;
import org.openjdk.jmh.annotations.Warmup;

import java.util.concurrent.TimeUnit;

// TODO: The parameters are way too low. Use for debugging only!
/*
 * Benchmark parameters
 */
@Fork(0)
@Warmup(iterations = 0)
@Measurement(iterations = 1)
@BenchmarkMode(Mode.AverageTime)
@OutputTimeUnit(TimeUnit.MILLISECONDS)
/*
 * Benchmark state parameters
 *
 * Quote from JMH:
 * State objects naturally encapsulate the state on which benchmark is working on.
 */
@State(Scope.Benchmark)
public abstract class AbstractBenchmark {
    @Param()
    private DataEnum dataEnum;

    private DataEnum prevDataEnum = null;
    private ObjectSupplier data;

    /* package-protected */ Object[] workingCopy;

    // TODO: This is inaccurate. How to create and use separate arrays for each warmup x iteration x sortAlgorithm ?
    @Setup(Level.Invocation)
    public void setup() {
        if (dataEnum != prevDataEnum){
            System.out.println(prevDataEnum + " => " + dataEnum);
            prevDataEnum = dataEnum;
            data = dataEnum.get();
        }
        workingCopy = data.getCopy();
    }

    @Benchmark
    public void benchmark() {
        sort();
    }

    /**
     * Sorts this.array with some sorting algorithm.
     */
    /* package-protected */ abstract void sort();
}
